import json
from os import getenv
from dotenv import load_dotenv
from os.path import isfile, isdir, join
from datetime import datetime, timezone, timedelta

from .davis import DavisParam, DavisReadings
from .spectrum import SpectrumParam, SpectrumReadings
from .zentra import ZentraParam, ZentraReadings
from .onset import OnsetParam, OnsetReadings
from .rainwise import RainwiseParam, RainwiseReadings
from .campbell import CampbellParam, CampbellReadings

__version__ = '0.0.21'


class ApiWrapper:
    def __init__(self, params: dict):
        self.vendor_list = ['zentra', 'spectrum', 'davis', 'onset', 'rainwise', 'campbell']
        self.vendor = params.get('vendor', None)
        self.params = params
        self.resp_raw = None
        self.resp_transformed = None
        self.resp_debug = None

        self.check_params()

    def check_params(self):
        if self.vendor is None or self.vendor.lower() not in self.vendor_list:
            raise Exception('"vendor" must be specified and in the approved vendor list.')
        else:
            self.vendor = self.vendor.lower()

    def get_reading(self):
        if self.vendor == 'zentra':
            zparam = ZentraParam(sn=self.params.get('sn', None),
                                 token=self.params.get('token', None),
                                 start_datetime=self.params.get('start_datetime', None),
                                 end_datetime=self.params.get('end_datetime', None),
                                 start_mrid=self.params.get('start_mrid', None),
                                 end_mrid=self.params.get('end_mrid', None),
                                 tz=self.params.get('tz', None),
                                 binding_ver=__version__)
            zreadings = ZentraReadings(zparam)
            self.resp_raw = zreadings.response
            self.resp_transformed = zreadings.transformed_resp
            self.resp_debug = zreadings.debug_info
            return self
        elif self.vendor == 'davis':
            dparam = DavisParam(sn=self.params.get('sn', None),
                                apikey=self.params.get('apikey', None),
                                apisec=self.params.get('apisec', None),
                                start_datetime=self.params.get('start_datetime', None),
                                end_datetime=self.params.get('end_datetime', None),
                                tz=self.params.get('tz', None),
                                binding_ver=__version__)
            dreadings = DavisReadings(dparam)
            self.resp_raw = dreadings.response
            self.resp_transformed = dreadings.transformed_resp
            self.resp_debug = dreadings.debug_info
            return self
        elif self.vendor == 'spectrum':
            sparam = SpectrumParam(sn=self.params.get('sn', None),
                                   apikey=self.params.get('apikey', None),
                                   start_datetime=self.params.get('start_datetime', None),
                                   end_datetime=self.params.get('end_datetime', None),
                                   date=self.params.get('date', None),
                                   tz=self.params.get('tz', None),
                                   count=self.params.get('count', None),
                                   binding_ver=__version__)
            sreadings = SpectrumReadings(sparam)
            self.resp_raw = sreadings.response
            self.resp_transformed = sreadings.transformed_resp
            self.resp_debug = sreadings.debug_info
            return self
        elif self.vendor == 'onset':
            oparam = OnsetParam(sn=self.params.get('sn', None),
                                client_id=self.params.get('client_id', None),
                                client_secret=self.params.get('client_secret', None),
                                ret_form=self.params.get('ret_form', None),
                                user_id=self.params.get('user_id', None),
                                start_datetime=self.params.get('start_datetime', None),
                                end_datetime=self.params.get('end_datetime', None),
                                tz=self.params.get('tz', None),
                                sensor_sn=self.params.get('sensor_sn', None),
                                binding_ver=__version__)
            oreadings = OnsetReadings(oparam)
            self.resp_raw = oreadings.response
            self.resp_transformed = oreadings.transformed_resp
            self.resp_debug = oreadings.debug_info
            return self
        elif self.vendor == 'rainwise':
            rparam = RainwiseParam(username=self.params.get('username', None),
                                   sid=self.params.get('sid', None),
                                   pid=self.params.get('pid', None),
                                   mac=self.params.get('mac', None),
                                   ret_form=self.params.get('ret_form', None),
                                   interval=self.params.get('interval', None),
                                   start_datetime=self.params.get('start_datetime', None),
                                   end_datetime=self.params.get('end_datetime', None),
                                   tz=self.params.get('tz', None),
                                   binding_ver=__version__)
            rreadings = RainwiseReadings(rparam)
            self.resp_raw = rreadings.response
            self.resp_transformed = rreadings.transformed_resp
            self.resp_debug = rreadings.debug_info
            return self
        elif self.vendor == 'campbell':
            cparam = CampbellParam(username=self.params.get('username', None),
                                   user_passwd=self.params.get('user_passwd', None),
                                   station_id=self.params.get('station_id', None),
                                   station_lid=self.params.get('station_lid', None),
                                   start_datetime=self.params.get('start_datetime', None),
                                   end_datetime=self.params.get('end_datetime', None),
                                   tz=self.params.get('tz', None),
                                   binding_ver=__version__)
            creadings = CampbellReadings(cparam)
            self.resp_raw = creadings.response
            self.resp_transformed = creadings.transformed_resp
            self.resp_debug = creadings.debug_info
            return self


def get_reading(vendor: str, **params) -> json:
    params['vendor'] = vendor
    return ApiWrapper(params).get_reading()


def get_version() -> str:
    return __version__


def get_sample_reports(out_dir, start_datetime=None, end_datetime=None):
    # if not env_file or not isfile(env_file):
    #     raise Exception("key file must be specified and/or key file does not exist")
    load_dotenv()
    if not out_dir or not isdir(out_dir):
        raise Exception("out_dir must be specified and/or out_dir folder does not exist")
    if start_datetime and not isinstance(start_datetime, datetime):
        raise Exception('start_datetime must be datetime.datetime instance')
    if end_datetime and not isinstance(end_datetime, datetime):
        raise Exception('end_datetime must be datetime.datetime instance')
    if start_datetime and end_datetime and (start_datetime > end_datetime):
        raise Exception('start_datetime must be earlier than end_datetime')
    if not start_datetime or not end_datetime:
        start_datetime = datetime.now() - timedelta(hours=24)
        end_datetime = start_datetime + timedelta(hours=2)

    def gen_report(vendor):
        params = json.loads(getenv(vendor.upper()))
        params['start_datetime'] = start_datetime
        params['end_datetime'] = end_datetime
        resp = get_reading(vendor, **params)
        with open(join(out_dir, vendor+'.json'), 'w') as wf:
            json.dump(resp.resp_raw, wf, indent=2)
        print(resp.resp_raw)

    # Campbell - Good
    gen_report('campbell')

    # Davis - Good
    gen_report('davis')

    # Onset - Good
    gen_report('onset')

    # Rainwise - Warning
    gen_report('rainwise')

    # Spectrum - Good
    gen_report('spectrum')

    # Zentra - Good
    gen_report('zentra')



from datetime import datetime, timezone
import json
import pytz
from requests import Session, Request


class OnsetParam:
    """
    A class used to represent Onset API parameters
    Attributes
    ----------
    sn : str
        The serial number of the device
    client_id : str
        client specific value provided by Onset
    client_secret : str
        client specific value provided by Onset
    ret_form : str
        The format data should be returned in. Currently only JSON is supported.
    user_id : str
        numeric ID of the user account This can be pulled from the HOBOlink URL: www.hobolink.com/users/<user_id>
    start_datetime_org : datetime
        Stores datetime object passed initially
    start_datetime : datetime
        Return readings with timestamps ≥ start_time. Specify start_time in Python Datetime format
    end_datetime_org : datetime
        Stores datetime object passed initially
    end_datetime : datetime
        Return readings with timestamps ≤ end_time. Specify end_time in Python Datetime format
    conversion_msg : str
        Stores time conversion message
    sensor_sn : dict
        a dict of sensor serial numbers
    json_file : str, optional
        The path to a local json file to parse.
    binding_ver : str
        Python binding version
    """
    def __init__(self, sn=None, client_id=None, client_secret=None, ret_form=None, user_id=None, start_datetime=None,
                 end_datetime=None, tz=None, sensor_sn=None, json_file=None, binding_ver=None):
        self.sn = sn
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = None
        self.ret_form = ret_form
        self.user_id = user_id
        self.path_param = None
        self.start_datetime_org = start_datetime
        self.start_datetime = start_datetime
        self.end_datetime_org = end_datetime
        self.end_datetime = end_datetime
        self.cur_datetime = datetime.now(timezone.utc)
        self.tz = tz
        self.conversion_msg = ''
        self.sensor_sn_dict = sensor_sn
        self.json_file = json_file
        self.binding_ver = binding_ver

        self.__check_params()
        self.__format_time()
        self.__get_auth()

    def __check_params(self):
        if self.start_datetime and not isinstance(self.start_datetime, datetime):
            raise Exception('start_datetime must be datetime.datetime instance')
        if self.end_datetime and not isinstance(self.end_datetime, datetime):
            raise Exception('end_datetime must be datetime.datetime instance')
        if self.start_datetime and self.end_datetime and (self.start_datetime > self.end_datetime):
            raise Exception('start_datetime must be earlier than end_datetime')
        if not self.json_file and not (self.start_datetime and self.end_datetime):
            raise Exception('state_datetime and end_datetime must be specified')
        if self.client_id is None or self.client_secret is None:
            raise Exception('client_id and client_secret parameters must be included')
        if self.ret_form is None or self.ret_form != 'JSON':
            raise Exception('ret_form must be specified and currently only \'JSON\' is supported')
        if self.user_id is None or not isinstance(self.user_id, str):
            raise Exception('userId must be specified and only str type is supported')
        self.path_param = {'format': self.ret_form, 'userId': self.user_id}

    def __utc_to_local(self):
        tzlist = {
            'HT': 'US/Hawaii',
            'AT': 'US/Alaska',
            'PT': 'US/Pacific',
            'MT': 'US/Mountain',
            'CT': 'US/Central',
            'ET': 'US/Eastern'
        }
        print('UTC Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'UTC start date passed as parameter: {}'.format(self.start_datetime) + " \\ "
        self.conversion_msg += 'Onset utilizes UTC timestamp as is, just added explicit UTC timezone' + " \\ "
        # self.start_datetime = self.start_datetime.replace(tzinfo=timezone.utc).astimezone(tz=None) \
        #     if self.start_datetime else None
        self.start_datetime = self.start_datetime.replace(tzinfo=timezone.utc) if self.start_datetime else None
        print('Explicit UTC time Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'Explicit UTC time start date after conversion: {}'.format(self.start_datetime) + " \\ "

        print('UTC End date: {}'.format(self.end_datetime))
        self.conversion_msg += 'UTC end date passed as parameter: {}'.format(self.end_datetime) + " \\ "
        # self.end_datetime = self.end_datetime.replace(tzinfo=timezone.utc).astimezone(tz=None)
        # if self.end_datetime else None
        self.end_datetime = self.end_datetime.replace(tzinfo=timezone.utc) if self.end_datetime else None
        self.conversion_msg += 'Explicit UTC time end date after conversion: {}'.format(self.end_datetime) + " \\ "
        print('Explicit UTC time End date: {}'.format(self.end_datetime))
        # for the metadata
        self.start_datetime_org = \
            self.start_datetime_org.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
        self.end_datetime_org = \
            self.end_datetime_org.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
        self.cur_datetime = self.cur_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))

    def __format_time(self):
        self.__utc_to_local()
        # self.start_datetime = self.start_datetime.astimezone(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
        # if self.start_datetime \
        self.start_datetime = self.start_datetime.strftime('%Y-%m-%d %H:%M:%S') if self.start_datetime \
            else datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
        self.end_datetime = self.end_datetime.strftime('%Y-%m-%d %H:%M:%S') if self.end_datetime \
            else datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
        # for metadata
        self.start_datetime_org = self.start_datetime_org.strftime('%Y-%m-%d %H:%M:%S')
        self.end_datetime_org = self.end_datetime_org.strftime('%Y-%m-%d %H:%M:%S')
        self.cur_datetime = self.cur_datetime.strftime('%Y-%m-%d %H:%M:%S')

    def __get_auth(self):
        print('client_id: \"{}\"'.format(self.client_id))
        print('client_secret: \"{}\"'.format(self.client_secret))
        request = Request('POST',
                          url='https://webservice.hobolink.com/ws/auth/token',
                          headers={
                              'Content-Type': 'application/x-www-form-urlencoded'},
                          data={'grant_type': 'client_credentials',
                                'client_id': self.client_id,
                                'client_secret': self.client_secret}).prepare()
        resp = Session().send(request)
        if resp.status_code != 200:
            raise Exception(
                'Get Auth request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code,
                                                                                             resp.text))
        response = resp.json()
        print('Auth response is \"{}\"'.format(response))
        self.access_token = response['access_token']
        print('access_token: \"{}\"'.format(self.access_token))


class OnsetReadings:
    """
    A class used to represent a device's readings
    Attributes
    ----------
    request : Request
        a Request object defining the request made to the Onset server
    response : Response
        a raw json response from the Onset server combined with meta data
    transformed_resp : list of dict
        a transformed response from raw JSON file or raw JSON response
    debug_info : dict
        a dict structure consist of parameter name and values
    """

    def __init__(self, param: OnsetParam):
        """
        Gets a device readings using a GET request to the Onset API.
        Parameters
        ----------
        param : OnsetParam
            OnsetParam object that contains Onset API parameters
        """
        self.debug_info = {
            'sn': param.sn,
            'access_token': param.access_token,
            'path_param': param.path_param,
            'start_datetime_org': param.start_datetime_org,
            'start_datetime': param.start_datetime,
            'end_datetime_org': param.end_datetime_org,
            'end_datetime': param.end_datetime,
            'cur_datetime': param.cur_datetime,
            'tz': param.tz,
            'conversion_msg': param.conversion_msg,
            'sensor_sn': param.sensor_sn_dict,
            'json_str': param.json_file,
            'binding_ver': param.binding_ver
        }
        if param.json_file:
            self.response = json.load(open(param.json_file))
            self.__transform()
        elif param.sn and param.access_token:
            self.__get(param.sn, param.access_token, param.path_param, param.start_datetime, param.end_datetime)
        elif param.sn or param.access_token:
            raise Exception('"sn" and "access_token" parameters must both be included.')
        else:
            # build an empty OnsetToken
            self.request = None
            self.response = None
            self.transformed_resp = None
            # self.device_info = None
            # self.measurement_settings = None
            # self.time_settings = None
            # self.locations = None
            # self.installation_metadata = None

    def __get(self, sn, access_token, path_param, start_datetime, end_datetime):
        """
        Gets a device readings using a GET request to the Onset API.
        Wraps build and parse functions.
        Parameters
        ----------
        sn : str
            The serial number of the device
        access_token : str
            The user's access token generated by OAuth protocol
        path_param : dict
            Path parameters used to formulate the endpoint during runtime
        start_datetime : str
            Return readings with timestamps ≥ start_datetime.
        end_datetime : str
            Return readings with timestamps ≤ end_datetime.
        """
        self.__build(sn, access_token, path_param, start_datetime, end_datetime)
        self.__make_request()
        self.__transform()
        return self

    def __build(self, sn, access_token, path_param, start_datetime, end_datetime):
        """
        Gets a device readings using a GET request to the Onset API.
        Parameters
        ----------
        sn : str
            The serial number of the device
        access_token : str
            The user's access token generated by OAuth protocol
        path_param : dict
            Path parameters used to formulate the endpoint during runtime
        start_datetime : str
            Return readings with timestamps ≥ start_datetime.
        end_datetime : str
            Return readings with timestamps ≤ end_datetime.
        """
        self.request = Request('GET',
                               url='https://webservice.hobolink.com/ws/data/file/{format}/user/{userId}'.format(
                                   **path_param),
                               headers={
                                   'Authorization': "Bearer " + access_token},
                               params={'loggers': sn,
                                       'start_date_time': start_datetime,
                                       'end_date_time': end_datetime}).prepare()
        self.debug_info['http_method'] = self.request.method
        self.debug_info['url'] = self.request.url
        self.debug_info['headers'] = self.request.headers
        return self

    def __make_request(self):
        """
        Sends a token request to the Onset API and stores the response.
        """
        # prep response list
        self.response = list()
        metadata = {
            "vendor": "onset",
            "station_id": self.debug_info['sn'],
            "timezone": self.debug_info['tz'],
            "start_datetime": self.debug_info['start_datetime_org'],
            "end_datetime": self.debug_info['end_datetime_org'],
            "request_time": self.debug_info['cur_datetime'],
            "python_binding_version": self.debug_info['binding_ver']}
        self.response.append(metadata)
        # Send the request and get the JSON response
        resp = Session().send(self.request)
        if resp.status_code != 200:
            raise Exception(
                'Request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code, resp.text))
        elif str(resp.content) == str(b'{"Error": "Device serial number entered does not exist"}'):
            raise Exception(
                'Error: Device serial number entered does not exist')
        self.response.append(resp.json())
        self.debug_info['response'] = self.response
        return self

    def __transform(self):
        """
        Parses the response.
        """
        def search_timestamp(input_datetime):
            if self.transformed_resp is None:
                return None
            if len(self.transformed_resp) == 0:
                return -1
            for x in range(len(self.transformed_resp)):
                if input_datetime == self.transformed_resp[x]['data_datetime']:
                    return x
            return -1

        def insert_resp(key, value, rec_datetime):
            resp_index = search_timestamp(rec_datetime)
            if resp_index is None:
                raise Exception("transformed_resp is None")
            elif resp_index == -1:
                temp_dict = {
                    "station_id": station_id,
                    "request_datetime": request_datetime,
                    "data_datetime": data_datetime,
                    key: value
                }
                self.transformed_resp.append(temp_dict)
            else:
                self.transformed_resp[resp_index][key] = value

        self.transformed_resp = list()
        station_id = self.response[0]['station_id']
        request_datetime = self.response[0]['request_time']
        for idx in range(1, len(self.response)):
            observe_list = self.response[idx]['observation_list']
            for kdx in range(len(observe_list)):
                if self.debug_info['sensor_sn']['atemp'] == observe_list[kdx]['sensor_sn']:
                    data_datetime = observe_list[kdx]['timestamp']
                    atemp = round(float(observe_list[kdx]['si_value']), 1)
                    insert_resp("atemp", atemp, data_datetime)

                if self.debug_info['sensor_sn']['pcpn'] == observe_list[kdx]['sensor_sn']:
                    data_datetime = observe_list[kdx]['timestamp']
                    pcpn = observe_list[kdx]['si_value']
                    insert_resp("pcpn", pcpn, data_datetime)

                if self.debug_info['sensor_sn']['relh'] == observe_list[kdx]['sensor_sn']:
                    data_datetime = observe_list[kdx]['timestamp']
                    relh = observe_list[kdx]['us_value']
                    insert_resp("relh", relh, data_datetime)

        print(self.transformed_resp)
        return self



import json
import pytz
from requests import Session, Request
from datetime import datetime, timezone


class SpectrumParam:
    """
    A class used to represent Spectrum API parameters
    Attributes
    ----------
    sn : str
        The serial number of the device
    apikey : str
        The customer's access key
    start_datetime_org : datetime
        Stores datetime object passed initially
    start_datetime : datetime
        Return readings for a specific customer device for a specific date-time range.
        (e.g., 2021-08-01 00:00)
    end_datetime_org : datetime
        Stores datetime object passed initially
    end_datetime : datetime
        Return readings for a specific customer device for a specific date-time range.
        (e.g., 2021-08-31 23:59)
    date_org : datetime
        Stores datetime object passed initially
    date : datetime
        Return readings for a specific date. (e.g., 2021-08-01)
    conversion_msg : str
        Stores time conversion message
    tz : str
        Time zone information
    count : int
        Get a specific number of recent sensor data records for a specific customer device
    json_file : str, optional
        The path to a local json file to parse.
    binding_ver : str
        Python binding version
    """
    def __init__(self, sn=None, apikey=None, start_datetime=None, end_datetime=None, date=None, tz=None, count=None,
                 json_file=None, binding_ver=None):
        self.sn = sn
        self.apikey = apikey
        self.start_datetime_org = start_datetime
        self.start_datetime = start_datetime
        self.end_datetime_org = end_datetime
        self.end_datetime = end_datetime
        self.date_org = date
        self.date = date
        self.cur_datetime = datetime.now(timezone.utc)
        self.tz = tz
        self.conversion_msg = ''
        self.count = count
        self.json_file = json_file
        self.binding_ver = binding_ver

        self.__check_params()
        self.__utc_to_local()

    def __check_params(self):
        tz_option = ['HT', 'AT', 'PT', 'MT', 'CT', 'ET']
        if self.start_datetime and not isinstance(self.start_datetime, datetime):
            raise Exception('start_datetime must be datetime.datetime instance')
        if self.end_datetime and not isinstance(self.end_datetime, datetime):
            raise Exception('end_datetime must be datetime.datetime instance')
        if self.start_datetime and self.end_datetime and (self.start_datetime > self.end_datetime):
            raise Exception('start_datetime must be earlier than end_datetime')
        if self.date and not isinstance(self.date, datetime):
            raise Exception('date must be datetime.datetime instance')
        if not self.json_file and not self.date and not (self.start_datetime and self.end_datetime):
            raise Exception('date or state_datetime and end_datetime must be specified')
        if self.tz and (self.tz not in tz_option):
            raise Exception('time zone options: HT, AT, PT, MT, CT, ET')
        if (self.start_datetime or self.end_datetime or self.date) and not self.tz:
            raise Exception('if start_datetime or end_datetime is specified, tz must be specified')

    def __utc_to_local(self):
        tzlist = {
            'HT': 'US/Hawaii',
            'AT': 'US/Alaska',
            'PT': 'US/Pacific',
            'MT': 'US/Mountain',
            'CT': 'US/Central',
            'ET': 'US/Eastern'
        }
        print('UTC Start date: {}, local time zone: {}'.format(self.start_datetime, self.tz))
        self.conversion_msg += \
            'UTC start date passed as parameter: {}, local time zone: {}'.format(self.start_datetime, self.tz) + " \\ "
        self.start_datetime = None if not self.start_datetime else \
            self.start_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
        # Spectrum API performs weird if time zone information is baked in the datetime object
        self.start_datetime = self.start_datetime.replace(tzinfo=None) if self.start_datetime else None
        print('Local time Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'Local time start date after conversion: {}'.format(self.start_datetime) + " \\ "

        print('UTC End date: {}, local time zone: {}'.format(self.end_datetime, self.tz))
        self.conversion_msg += \
            'UTC end date passed as parameter: {}, local time zone: {}'.format(self.end_datetime, self.tz) + " \\ "
        self.end_datetime = None if not self.end_datetime else \
            self.end_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
        # Spectrum API performs weird if time zone information is baked in the datetime object
        self.end_datetime = self.end_datetime.replace(tzinfo=None) if self.end_datetime else None
        self.conversion_msg += 'Local time end date after conversion: {}'.format(self.end_datetime) + " \\ "
        print('Local time End date: {}'.format(self.end_datetime))

        if self.date:
            print('UTC date: {}, local time zone: {}'.format(self.date, self.tz))
            self.conversion_msg += \
                'UTC date passed as parameter: {}, local time zone: {}'.format(self.date, self.tz) + " \\ "
            self.date = self.date.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
            # Spectrum API performs weird if time zone information is baked in the datetime object
            self.date = self.date.replace(tzinfo=None)
            self.conversion_msg += 'Local time date after conversion: {}'.format(self.date) + " \\ "
            print('Local time End date: {}'.format(self.date))

        self.cur_datetime = self.cur_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))


class SpectrumReadings:
    """
    A class used to represent a device's readings
    Attributes
    ----------
    request : Request
        a Request object defining the request made to the Spectrum server
    response : Response
        a json response from the Spectrum server
    transformed_resp : list of dict
        a transformed response from raw JSON file or raw JSON response
    debug_info : dict
        a dict structure consist of parameter name and values
    """
    def __init__(self, param: SpectrumParam):
        """
        Gets a device readings using a GET request to the Spectrum API.

        Parameters
        ----------
        param: SpectrumParam
            SpectrumParam object that contains Spectrum API parameters
        """
        self.debug_info = {
            'sn': param.sn,
            'apikey': param.apikey,
            'start_datetime_org': param.start_datetime_org,
            'start_datetime': param.start_datetime,
            'end_datetime_org': param.end_datetime_org,
            'end_datetime': param.end_datetime,
            'date_org': param.date_org,
            'date': param.date,
            'cur_datetime': param.cur_datetime.strftime('%Y-%m-%d %H:%M:%S'),
            'tz': param.tz,
            'conversion_msg': param.conversion_msg,
            'count': param.count,
            'json_str': param.json_file,
            'binding_ver': param.binding_ver
        }
        if param.json_file:
            self.response = json.load(open(param.json_file))
            self.__transform()
        elif param.sn and param.apikey:
            self.__get(param.sn, param.apikey, param.start_datetime, param.end_datetime, param.date, param.count)
        elif param.sn or param.apikey:
            raise Exception('"sn" and "apikey" parameters must both be included.')
        else:
            # build an empty Spectrum apikey
            self.request = None
            self.response = None
            self.transformed_resp = None
            # self.device_info = None
            # self.measurement_settings = None
            # self.time_settings = None
            # self.locations = None
            # self.installation_metadata = None

    def __get(self, sn, apikey, start_datetime=None, end_datetime=None, date=None, count=None):
        """
        Gets a device readings using a GET request to the Spectrum API.
        Wraps build and parse functions.
        Parameters
        ----------
        sn : str
            The serial number of the device
        apikey : Spectrum apikey
            The user's access apikey
        start_datetime : datetime
            Return readings for a specific customer device for a specific date/time range.
            (e.g., 2021-08-01, 2021-08-01 00:00)
        end_datetime : datetime
            Return readings for a specific customer device for a specific date/time range.
            (e.g., 2021-08-31, 2021-08-31 23:59)
        date : datetime
            Return readings for a specific date. (e.g., 2021-08-01)
        count : int
            Get a specific number of recent sensor data records for a specific customer device
        """
        self.__build(sn, apikey, start_datetime, end_datetime, date, count)
        self.__make_request()
        self.__transform()
        return self

    def __build(self, sn, apikey, start_datetime=None, end_datetime=None, date=None, count=None):
        """
        Gets a device readings using a GET request to the Spectrum API.
        Parameters
        ----------
        sn : str
            The serial number of the device
        apikey : Spectrum apikey
            The user's access apikey
        start_datetime : datetime
            Return readings for a specific customer device for a specific date/time range.
            (e.g., 2021-08-01, 2021-08-01 00:00)
        end_datetime : datetime
            Return readings for a specific customer device for a specific date/time range.
            (e.g., 2021-08-31, 2021-08-31 23:59)
        date : datetime
            Return readings for a specific date. (e.g., 2021-08-01)
        count : int
            Get a specific number of recent sensor data records for a specific customer device
        """
        if count and count > 0:
            self.request = Request('GET',
                                   url='https://api.specconnect.net:6703/api/Customer/GetDataByRange',
                                   params={'customerApiKey': apikey, 'serialNumber': sn, 'count': count}).prepare()
        elif date:
            self.request = Request('GET',
                                   url='https://api.specconnect.net:6703/api/Customer/GetDataByDate',
                                   params={'customerApiKey': apikey, 'serialNumber': sn, 'date': date}).prepare()
        elif start_datetime and end_datetime:
            self.request = Request('GET',
                                   url='https://api.specconnect.net:6703/api/Customer/GetDataInDateTimeRange',
                                   params={'customerApiKey': apikey, 'serialNumber': sn,
                                           'startDate': start_datetime, 'endDate': end_datetime}).prepare()
        else:
            self.request = Request('GET',
                                   url='https://api.specconnect.net:6703/api/Customer/GetData',
                                   params={'customerApiKey': apikey, 'serialNumber': sn}).prepare()

        self.debug_info['http_method'] = self.request.method
        self.debug_info['url'] = self.request.url
        self.debug_info['headers'] = self.request.headers
        return self

    def __make_request(self):
        """
        Sends a token request to the Spectrum API and stores the response.
        """
        # prep response list
        self.response = list()
        metadata = {
            "vendor": "spectrum",
            "station_id": self.debug_info['sn'],
            "start_datetime": self.debug_info['start_datetime'].strftime('%Y-%m-%d %H:%M:%S') if
            self.debug_info['start_datetime'] else None,
            "end_datetime": self.debug_info['end_datetime'].strftime('%Y-%m-%d %H:%M:%S') if
            self.debug_info['end_datetime'] else None,
            "timezone": self.debug_info['tz'],
            "request_time": self.debug_info['cur_datetime'],
            "python_binding_version": self.debug_info['binding_ver']}

        self.response.append(metadata)
        # Send the request and get the JSON response
        resp = Session().send(self.request)
        if resp.status_code != 200:
            raise Exception(
                'Request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code, resp.text))
        self.response.append(resp.json())
        self.debug_info['response'] = self.response
        return self

    def __transform(self):
        """
        Parses the response.
        """
        self.transformed_resp = list()
        station_id = self.response[0]['station_id']
        request_datetime = self.response[0]['request_time']
        for idx in range(1, len(self.response)):
            equipment_rec = self.response[idx]['EquipmentRecords']
            for jdx in range(len(equipment_rec)):
                sensor_data = equipment_rec[jdx]['SensorData']
                temp_dict = {
                    "station_id": station_id,
                    "request_datetime": request_datetime,
                }
                for kdx in range(len(sensor_data)):
                    if sensor_data[kdx]['SensorType'] == 'Temperature':
                        temp_dict['data_datetime'] = sensor_data[kdx]['FormattedTimeStamp']
                        temp_dict['atemp'] = round(((float(sensor_data[kdx]['Value'])-32)*5/9), 1)
                        # print(temp_dict['atemp'])
                    if sensor_data[kdx]['SensorType'] == 'Rainfall':
                        temp_dict['pcpn'] = (float(sensor_data[kdx]['Value'])*25.4)
                    if sensor_data[kdx]['SensorType'] == 'Relative Humidity':
                        temp_dict['relh'] = sensor_data[kdx]['Value']
                self.transformed_resp.append(temp_dict)
        print(self.transformed_resp)
        return self



from datetime import datetime, timezone
import json
import pytz
from requests import Session, Request


class RainwiseParam:
    """
    A class used to represent Rainwise API parameters
    Attributes
    ----------
    username : str
        Registered group name (same as MAC)
    sid : str
        Site id, assigned by Rainwise
    pid : str
        Password id, assigned by Rainwise
    mac : str
        MAC of the weather station, Must be in the group assigned to username
    ret_form: str, optional (default json)
        Values xml or json; returns the data as JSON or XML
    interval: int, optional (default 1 min)
        Data aggregation interval, 1, 5, 10, 15, 30, 60 minute intervals
    start_datetime_org : datetime
        Stores datetime object passed initially
    start_datetime : datetime (UTC expected)
        Return readings with timestamps ≥ start_time. Specify start_time in Python Datetime format
    end_datetime_org : datetime
        Stores datetime object passed initially
    end_datetime : datetime (UTC expected)
        Return readings with timestamps ≤ end_time. Specify end_time in Python Datetime format
    tz : str
        Time zone information
    conversion_msg : str
        Stores time conversion message
    json_file : str, optional
        The path to a local json file to transform
    binding_ver : str
        Python binding version
    """
    def __init__(self, username=None, sid=None, pid=None, mac=None, ret_form='json', interval=1,
                 start_datetime=None, end_datetime=None, tz=None, json_file=None, binding_ver=None):
        self.username = username
        self.sid = sid
        self.pid = pid
        self.mac = mac
        self.ret_form = ret_form
        self.interval = interval
        self.start_datetime_org = start_datetime
        self.start_datetime = start_datetime
        self.end_datetime_org = end_datetime
        self.end_datetime = end_datetime
        self.cur_datetime = datetime.now(timezone.utc)
        self.tz = tz
        self.conversion_msg = ''

        self.json_file = json_file
        self.binding_ver = binding_ver

        self.__check_params()
        self.__format_time()

    def __check_params(self):
        tz_option = ['HT', 'AT', 'PT', 'MT', 'CT', 'ET']
        if self.start_datetime and not isinstance(self.start_datetime, datetime):
            raise Exception('start_datetime must be datetime.datetime instance')
        if self.end_datetime and not isinstance(self.end_datetime, datetime):
            raise Exception('end_datetime must be datetime.datetime instance')
        if self.start_datetime and self.end_datetime and (self.start_datetime > self.end_datetime):
            raise Exception('start_datetime must be earlier than end_datetime')
        if self.tz and (self.tz not in tz_option):
            raise Exception('time zone options: HT, AT, PT, MT, CT, ET')
        if not self.json_file and not (self.start_datetime and self.end_datetime):
            raise Exception('state_datetime and end_datetime must be specified')
        if (self.start_datetime or self.end_datetime) and not self.tz:
            raise Exception('if start_datetime or end_datetime is specified, tz must be specified')
        if self.username is None or self.mac is None or self.username != self.mac:
            raise Exception('username and mac parameters must be included and same value')
        if self.sid is None or self.pid is None or self.sid != self.pid:
            raise Exception('sid and pid parameters must be included and same value')
        if self.ret_form.lower() != 'json' and self.ret_form.lower() != 'xml':
            raise Exception('ret_form must either be json or xml')

    def __utc_to_local(self):
        tzlist = {
            'HT': 'US/Hawaii',
            'AT': 'US/Alaska',
            'PT': 'US/Pacific',
            'MT': 'US/Mountain',
            'CT': 'US/Central',
            'ET': 'US/Eastern'
        }
        print('UTC Start date: {}, local time zone: {}'.format(self.start_datetime, self.tz))
        self.conversion_msg += \
            'UTC start date passed as parameter: {}, local time zone: {}'.format(self.start_datetime, self.tz) + " \\ "
        # self.start_datetime = self.start_datetime.replace(tzinfo=timezone.utc).astimezone(tz=None) \
        #     if self.start_datetime else None
        self.start_datetime = \
            self.start_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz])) \
            if self.start_datetime else None
        print('Local time Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'Local time start date after conversion: {}'.format(self.start_datetime) + " \\ "

        print('UTC End date: {}, local time zone: {}'.format(self.end_datetime, self.tz))
        self.conversion_msg += \
            'UTC end date passed as parameter: {}, local time zone: {}'.format(self.end_datetime, self.tz) + " \\ "
        # self.end_datetime = self.end_datetime.replace(tzinfo=timezone.utc).astimezone(tz=None)
        # if self.end_datetime else None
        self.end_datetime = self.end_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz])) \
            if self.end_datetime else None
        print('Local time End date: {}'.format(self.end_datetime))
        self.conversion_msg += 'Local time end date after conversion: {}'.format(self.end_datetime) + " \\ "
        self.cur_datetime = self.cur_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))

    def __format_time(self):
        self.__utc_to_local()
        self.start_datetime = self.start_datetime.strftime('%Y-%m-%d %H:%M:%S') if self.start_datetime \
            else datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        self.end_datetime = self.end_datetime.strftime('%Y-%m-%d %H:%M:%S') if self.end_datetime \
            else datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        self.cur_datetime = self.cur_datetime.strftime('%Y-%m-%d %H:%M:%S')


class RainwiseReadings:
    """
    A class used to represent a device's readings
    Attributes
    ----------
    request : Request
        a Request object defining the request made to the Rainwise server
    response : list
        a raw json response from the Rainwise server combined with meta data
    transformed_resp : list of dict
        a transformed response from raw JSON file or raw JSON response
    """

    def __init__(self, param: RainwiseParam):
        """
        Gets a device readings using a GET request to the Rainwise API.
        Parameters
        ----------
        param : RainwiseParam
            RainwiseParam object that contains Rainwise API parameters
        """
        self.debug_info = {
            'username': param.username,
            'sid': param.sid,
            'pid': param.pid,
            'mac': param.mac,
            'ret_form': param.ret_form,
            'interval': param.interval,
            'start_datetime_org': param.start_datetime_org,
            'start_datetime': param.start_datetime,
            'end_datetime_org': param.end_datetime_org,
            'end_datetime': param.end_datetime,
            'cur_datetime': param.cur_datetime,
            'tz': param.tz,
            'conversion_msg': param.conversion_msg,
            'json_str': param.json_file,
            'binding_ver': param.binding_ver
        }
        if param.json_file:
            self.response = json.load(open(param.json_file))
            self.__transform()
        elif param.username and param.sid and param.pid and param.mac:
            self.__get(param.username, param.sid, param.pid, param.mac, param.ret_form, param.interval,
                       param.start_datetime, param.end_datetime)
        elif param.username or param.sid or param.pid or param.mac:
            raise Exception('"username", "sid", "pid", "mac" parameters must be included.')
        else:
            # build an empty RainwiseToken
            self.request = None
            self.response = None
            self.transformed_resp = None
            # self.device_info = None
            # self.measurement_settings = None
            # self.time_settings = None
            # self.locations = None
            # self.installation_metadata = None

    def __get(self, username, sid, pid, mac, ret_form, interval, start_datetime=None, end_datetime=None):
        """
        Gets a device readings using a GET request to the Rainwise API.
        Wraps build and transform functions.
        Parameters
        ----------
        username : str
            Registered group name (may be same as MAC)
        sid : str
            Site id, assigned by Rainwise
        pid : str
            Password id, assigned by Rainwise
        mac : str
            MAC of the weather station, Must be in the group assigned to username
        ret_form: str, optional (default json)
            Values xml or json; returns the data as JSON or XML
        interval: int, optional (default 1 min)
            Data aggregation interval, 1, 5, 10, 15, 30, 60 minute intervals
        start_datetime : datetime
            Return readings with timestamps ≥ start_time. Specify start_time in Python Datetime format
        end_datetime : datetime
            Return readings with timestamps ≤ end_time. Specify end_time in Python Datetime format
        """
        self.__build(username, sid, pid, mac, ret_form, interval, start_datetime, end_datetime)
        self.__make_request()
        self.__transform()
        return self

    def __build(self, username, sid, pid, mac, ret_form, interval, start_datetime=None, end_datetime=None):
        """
        Gets a device readings using a GET request to the Rainwise API.
        Parameters
        ----------
        username : str
            Registered group name (may be same as MAC)
        sid : str
            Site id, assigned by Rainwise
        pid : str
            Password id, assigned by Rainwise
        mac : str
            MAC of the weather station, Must be in the group assigned to username
        ret_form: str, optional (default json)
            Values xml or json; returns the data as JSON or XML
        interval: int, optional (default 1 min)
            Data aggregation interval, 1, 5, 10, 15, 30, 60 minute intervals
        start_datetime : datetime
            Return readings with timestamps ≥ start_time. Specify start_time in Python Datetime format
        end_datetime : datetime
            Return readings with timestamps ≤ end_time. Specify end_time in Python Datetime format
        """
        self.request = Request('GET',
                               url='http://api.rainwise.net/main/v1.5/registered/get-historical.php',
                               params={'username': username,
                                       'sid': sid,
                                       'pid': pid,
                                       'mac': mac,
                                       'format': ret_form,
                                       'interval': interval,
                                       'sdate': start_datetime,
                                       'edate': end_datetime}).prepare()

        self.debug_info['http_method'] = self.request.method
        self.debug_info['url'] = self.request.url
        self.debug_info['headers'] = self.request.headers
        return self

    def __make_request(self):
        """
        Sends a token request to the Rainwise API and stores the response.
        """
        # prep response list
        self.response = list()
        metadata = {
            "vendor": "rainwise",
            "station_id": self.debug_info['mac'],
            "timezone": self.debug_info['tz'],
            "start_datetime": self.debug_info['start_datetime'],
            "end_datetime": self.debug_info['end_datetime'],
            "request_time": self.debug_info['cur_datetime'],
            "python_binding_version": self.debug_info['binding_ver']}
        self.response.append(metadata)
        # Send the request and get the JSON response
        resp = Session().send(self.request)
        if resp.status_code != 200:
            raise Exception(
                'Request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code, resp.text))
        elif str(resp.content) == str(b'{"Error": "Device serial number entered does not exist"}'):
            raise Exception(
                'Error: Device serial number entered does not exist')
        self.response.append(resp.json())
        self.debug_info['response'] = self.response
        return self

    def __transform(self):
        """
        Transform the response.
        """
        self.transformed_resp = list()
        station_id = self.response[0]['station_id']
        request_datetime = self.response[0]['request_time']
        for idx in range(1, len(self.response)):
            for k, v in self.response[idx]['times'].items():
                temp_dic = {
                    "station_id": station_id,
                    "request_datetime": request_datetime,
                    "data_datetime": v,
                    "atemp": round(((float(self.response[idx]['temp'][k]) - 32) * 5 / 9), 1),
                    "pcpn": (float(self.response[idx]['precip'][k])*25.4),
                    "relh": self.response[idx]['hum'][k]
                }
                self.transformed_resp.append(temp_dic)
        print(self.transformed_resp)
        return self



from datetime import datetime, timezone
import json
import pytz
from requests import Session, Request


class CampbellParam:
    """
    A class used to represent Campbell API parameters
    Attributes
    ----------
    username : str
        Login account username to campbell-cloud
    user_passwd : str
        Login account password to campbell-cloud
    credentials : dict
        Username and Password dict (internally generated)
    access_token : str
        API access token generated using username & password (internally generated)
    kc_id : str
        Alphanumeric user ID from user details (internally generated)
    organization_id: str
        Alphanumeric organization_id  from user details (internally generated)
    measurements : str
        The name or names of available measurements of the station (comma separated str, internally generated)
    station_id : str
        Alphanumeric ID of the station (for v3 APIs)
    station_lid : str
        Alphanumeric Legacy ID of the station (for v2 APIs)
    start_datetime_org : datetime
        Stores datetime object passed initially
    start_datetime : datetime
        Return readings with timestamps ≥ start_time. Specify start_time in Python Datetime format
    end_datetime_org : datetime
        Stores datetime object passed initially
    end_datetime : datetime
        Return readings with timestamps ≤ end_time. Specify end_time in Python Datetime format
    conversion_msg : str
        Stores time conversion message
    json_file : str, optional
        The path to a local json file to transform
    binding_ver : str
        Python binding version
    """
    def __init__(self, username=None, user_passwd=None, station_id=None, station_lid=None,
                 start_datetime=None, end_datetime=None, tz=None, json_file=None, binding_ver=None):

        self.username = username
        self.user_passwd = user_passwd
        self.credentials = None
        self.access_token = None
        self.kc_id = None
        self.organization_id = None
        self.measurements = ''
        self.station_id = station_id
        self.station_lid = station_lid

        self.start_datetime_org = start_datetime
        self.start_datetime = start_datetime
        self.end_datetime_org = end_datetime
        self.end_datetime = end_datetime
        self.cur_datetime = datetime.now(timezone.utc)
        self.tz = tz
        self.conversion_msg = ''
        self.json_file = json_file
        self.binding_ver = binding_ver

        self.__check_params()
        self.__format_time()
        self.__get_auth()
        self.__get_ids()
        self.__get_measurements()

    def __check_params(self):
        if self.username is None or not isinstance(self.username, str):
            raise Exception('username must be specified and only str type is supported')
        if self.user_passwd is None or not isinstance(self.user_passwd, str):
            raise Exception('user_passwd must be specified and only str type is supported')
        self.credentials = {'username': self.username, 'password': self.user_passwd}
        if self.station_id is None or self.station_lid is None:
            raise Exception('station_id and station_lid parameters must be included')
        if self.start_datetime and not isinstance(self.start_datetime, datetime):
            raise Exception('start_datetime must be datetime.datetime instance')
        if self.end_datetime and not isinstance(self.end_datetime, datetime):
            raise Exception('end_datetime must be datetime.datetime instance')
        if self.start_datetime and self.end_datetime and (self.start_datetime > self.end_datetime):
            raise Exception('start_datetime must be earlier than end_datetime')
        if not self.json_file and not (self.start_datetime and self.end_datetime):
            raise Exception('state_datetime and end_datetime must be specified')
        print("user credentials: {} type - {}".format(type(self.credentials), self.credentials))

    def __utc_to_local(self):
        tzlist = {
            'HT': 'US/Hawaii',
            'AT': 'US/Alaska',
            'PT': 'US/Pacific',
            'MT': 'US/Mountain',
            'CT': 'US/Central',
            'ET': 'US/Eastern'
        }
        print('UTC Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'UTC start date passed as parameter: {}'.format(self.start_datetime) + " \\ "
        self.conversion_msg += 'Campbell utilizes Unix Epoch, just added explicit UTC timezone' + " \\ "
        # self.start_datetime = self.start_datetime.replace(tzinfo=timezone.utc).astimezone(tz=None) \
        #     if self.start_datetime else None
        self.start_datetime = self.start_datetime.replace(tzinfo=timezone.utc) if self.start_datetime else None
        print('Explicit UTC time Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'Explicit UTC time start date after conversion: {}'.format(self.start_datetime) + " \\ "

        print('UTC End date: {}'.format(self.end_datetime))
        self.conversion_msg += 'UTC end date passed as parameter: {}'.format(self.end_datetime) + " \\ "
        # self.end_datetime = self.end_datetime.replace(tzinfo=timezone.utc).astimezone(tz=None)
        # if self.end_datetime else None
        self.end_datetime = self.end_datetime.replace(tzinfo=timezone.utc) if self.end_datetime else None
        self.conversion_msg += 'Explicit UTC time end date after conversion: {}'.format(self.end_datetime) + " \\ "
        print('Explicit UTC time End date: {}'.format(self.end_datetime))
        # for the metadata
        self.start_datetime_org = \
            self.start_datetime_org.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
        self.end_datetime_org = \
            self.end_datetime_org.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
        self.cur_datetime = self.cur_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))

    def __format_time(self):
        self.__utc_to_local()
        self.start_datetime = int(self.start_datetime.timestamp()*1000) if self.start_datetime \
            else int(datetime.now().timestamp()*1000)
        self.end_datetime = int(self.end_datetime.timestamp()*1000) if self.end_datetime \
            else int(datetime.now().timestamp()*1000)
        # for metadata
        self.start_datetime_org = self.start_datetime_org.strftime('%Y-%m-%d %H:%M:%S')
        self.end_datetime_org = self.end_datetime_org.strftime('%Y-%m-%d %H:%M:%S')
        self.cur_datetime = self.cur_datetime.strftime('%Y-%m-%d %H:%M:%S')

    def __get_auth(self):
        print('username: \"{}\"'.format(self.username))
        print('user_password: \"{}\"'.format(self.user_passwd))
        request = Request('POST',
                          url='https://api.campbellcloud.io/v3/campbell-cloud/tokens',
                          headers={
                              'Content-Type': 'application/json'},
                          json={'grant_type': 'password',
                                'credentials': self.credentials}).prepare()
        resp = Session().send(request)
        if resp.status_code != 200:
            raise Exception(
                'Get Auth request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code,
                                                                                             resp.text))
        response = resp.json()
        print('Auth response is \"{}\"'.format(response))
        self.access_token = response['access_token']
        print('access_token: \"{}\"'.format(self.access_token))

    def __get_ids(self):
        request = Request('GET',
                          url='https://api.campbellcloud.io/api_v2/user/session',
                          headers={
                              'Authorization': 'Bearer ' + self.access_token
                          }).prepare()
        resp = Session().send(request)
        if resp.status_code != 200:
            raise Exception(
                'Get IDs request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code,
                                                                                            resp.text))
        response = resp.json()
        print('User details are \"{}\"'.format(response))
        self.kc_id = response['kc_id']
        self.organization_id = response['organization_id']
        print('kc_id is \"{}\"'.format(self.kc_id))
        print('organization_id is \"{}\"'.format(self.organization_id))

    def __get_measurements(self):
        path_param = {'organization_id': self.organization_id, 'station_id': self.station_id}
        request = Request('GET',
                          url='https://api.campbellcloud.io/v3/campbell-cloud/organizations'
                              '/{organization_id}/stations/{station_id}/definitions'.format(**path_param),
                          headers={
                              'Authorization': 'Bearer ' + self.access_token},
                          params={'brief': 'true'}).prepare()
        resp = Session().send(request)
        if resp.status_code != 200:
            raise Exception(
                'Get IDs request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code,
                                                                                            resp.text))
        response = resp.json()
        for measurement_data in response:
            self.measurements += measurement_data['name'] + ','
        self.measurements = self.measurements[:-1]
        print('Available measurements are \n{}'.format(self.measurements))


class CampbellReadings:
    """
    A class used to represent a device's readings
    Attributes
    ----------
    request : Request
        a Request object defining the request made to the Campbell server
    response : Response
        a json response from the Campbell server
    transformed_resp : list of dict
        a transformed response from raw JSON file or raw JSON response
    debug_info : dict
        a dict structure consist of parameter name and values
    """

    def __init__(self, param: CampbellParam):
        """
        Gets a device readings using a GET request to the Campbell API.
        Parameters
        ----------
        param : CampbellParam
            CampbellParam object that contains Campbell API parameters
        """
        self.debug_info = {
            'station_lid': param.station_lid,
            'start_datetime_org': param.start_datetime_org,
            'start_datetime': param.start_datetime,
            'end_datetime_org': param.end_datetime_org,
            'end_datetime': param.end_datetime,
            'cur_datetime': param.cur_datetime,
            'tz': param.tz,
            'conversion_msg': param.conversion_msg,
            'measurements': param.measurements,
            'access_token': param.access_token,
            'json_str': param.json_file,
            'binding_ver': param.binding_ver
        }
        if param.json_file:
            self.response = json.load(open(param.json_file))
            self.__transform()
        elif param.station_lid and param.access_token:
            self.__get(param.station_lid, param.start_datetime, param.end_datetime, param.measurements,
                       param.access_token)
        elif param.station_lid or param.access_token:
            raise Exception('"station_id" and "access_token" parameters must both be included.')
        else:
            # build an empty CampbellToken
            self.request = None
            self.response = None
            self.transformed_resp = None
            # self.device_info = None
            # self.measurement_settings = None
            # self.time_settings = None
            # self.locations = None
            # self.installation_metadata = None

    def __get(self, station_lid, epoch_start, epoch_end, measurements, access_token):
        """
        Gets a device readings using a GET request to the Campbell API.
        Wraps build and parse functions.
        Parameters
        ----------
        station_lid : str
            Alphanumeric Legacy ID of the station (for v2 APIs)
        epoch_start : int
            epoch timestamp generated for the beginning of your time period
        epoch_end : int
            epoch timestamp generated for the end of your time period
        measurements : str
            comma separated list of measurements
        access_token : str
            The user's access token for API authorization
        """
        self.__build(station_lid, epoch_start, epoch_end, measurements, access_token)
        self.__make_request()
        self.__transform()
        return self

    def __build(self, station_lid, epoch_start, epoch_end, measurements, access_token):
        """
        Gets a device readings using a GET request to the Campbell API.
        Parameters
        ----------
        station_lid : str
            Alphanumeric Legacy ID of the station (for v2 APIs)
        epoch_start : int
            epoch timestamp generated for the beginning of your time period
        epoch_end : int
            epoch timestamp generated for the end of your time period
        measurements : str
            comma separated list of measurements
        access_token : str
            The user's access token for API authorization
        """
        path_param = {
            'station_id': station_lid,
            'epoch_start': epoch_start,
            'epoch_end': epoch_end,
            'measurements': measurements
        }
        self.request = Request('GET',
                               url='https://api.campbellcloud.io/api_v2/measurement/timeseries'
                                   '/{station_id}/{epoch_start}/{epoch_end}/{measurements}'.format(**path_param),
                               headers={
                                   'Authorization': "Bearer " + access_token}).prepare()
        self.debug_info['http_method'] = self.request.method
        self.debug_info['url'] = self.request.url
        self.debug_info['headers'] = self.request.headers
        return self

    def __make_request(self):
        """
        Sends a token request to the Campbell API and stores the response.
        """
        # prep response list
        self.response = list()
        metadata = {
            "vendor": "campbell",
            "station_id": self.debug_info['station_lid'],
            "timezone": self.debug_info['tz'],
            "start_datetime": self.debug_info['start_datetime_org'],
            "end_datetime": self.debug_info['end_datetime_org'],
            "request_time": self.debug_info['cur_datetime'],
            "python_binding_version": self.debug_info['binding_ver']}
        self.response.append(metadata)
        # Send the request and get the JSON response
        resp = Session().send(self.request)
        if resp.status_code != 200:
            raise Exception(
                'Request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code, resp.text))
        elif str(resp.content) == str(b'{"Error": "Device serial number entered does not exist"}'):
            raise Exception(
                'Error: Device serial number entered does not exist')
        self.response.append(resp.json())
        self.debug_info['response'] = self.response
        return self

    def __transform(self):
        """
        Transform the response.
        """
        self.transformed_resp = list()
        # try:
        #     self.device_info = self.response['device']['device_info']
        # except KeyError:
        #     self.device_info = 'N/A'
        # self.timeseries = list(
        #     map(lambda x: CampbellTimeseriesRecord(x), self.response['device']['timeseries']))
        print(self.transformed_resp)
        return self



import collections
import hashlib
import hmac
import json
import pytz
from datetime import datetime, timezone, timedelta
from requests import Session, Request


class DavisParam:
    """
    A class used to represent Davis API parameters
    Attributes
    ----------
    sn : str
        The serial number of the device
    apikey : str
        The customer's access key (v2)
    apisec : str
        API security that is used to compute the hash
    start_datetime : datetime (UTC expected)
        Return readings with timestamps ≥ start_time. Specify start_time in str. (2021-08-01, 2021-08-01)
    end_datetime : datetime (UTC expected)
        Return readings with timestamps ≤ end_time. Specify end_time in str. (2021-08-31, 2021-08-31)
    date_tuple_list : list
        A list of date/time tuples with API signatures
    conversion_msg : str
        Stores time conversion message
    json_file : str, optional
        The path to a local json file to parse
    binding_ver : str
        Python binding version
    """
    def __init__(self, sn=None, apikey=None, apisec=None, start_datetime=None, end_datetime=None, tz=None,
                 json_file=None, binding_ver=None):
        self.sn = sn
        self.apikey = apikey
        self.apisec = apisec
        self.apisig = None  # used when start_datetime & end_datetime is passed
        self.t = int(datetime.now().timestamp())
        # self.start_datetime = int(time.mktime(time.strptime(start_datetime, "%m/%d/%Y %H:%M")))
        # if start_datetime else None
        # self.end_datetime = int(time.mktime(time.strptime(end_datetime, "%m/%d/%Y %H:%M"))) if end_datetime else None
        self.start_datetime = start_datetime
        self.end_datetime = end_datetime
        self.cur_datetime = datetime.now(timezone.utc)
        self.tz = tz
        self.date_tuple_list = list()
        self.conversion_msg = ''
        self.json_file = json_file
        self.binding_ver = binding_ver

        self.__check_params()
        self.__format_time()

        if json_file is None:
            self.__compute_signature()

    def __check_params(self):
        if self.start_datetime and not isinstance(self.start_datetime, datetime):
            raise Exception('start_datetime must be datetime.datetime instance')
        if self.end_datetime and not isinstance(self.end_datetime, datetime):
            raise Exception('end_datetime must be datetime.datetime instance')
        if self.start_datetime and self.end_datetime and (self.start_datetime > self.end_datetime):
            raise Exception('start_datetime must be earlier than end_datetime')
        if not self.json_file and not (self.start_datetime and self.end_datetime):
            raise Exception('state_datetime and end_datetime must be specified')
        if self.apikey is None or self.apisec is None:
            raise Exception('"apikey" and "apisec" parameters must both be included.')
        if self.sn is None:
            raise Exception('"sn" parameter must be included.')

        if self.start_datetime and self.end_datetime:
            # remove sub-second since API cannot distinguish them different
            self.start_datetime = self.start_datetime.replace(microsecond=0)
            self.end_datetime = self.end_datetime.replace(microsecond=0)
            if self.end_datetime - timedelta(hours=24) <= self.start_datetime:
                self.date_tuple_list.append((self.start_datetime, self.end_datetime))
            else:
                while self.end_datetime - timedelta(hours=24) > self.start_datetime:
                    self.date_tuple_list.append(
                        (self.start_datetime, self.start_datetime + timedelta(hours=23, minutes=59, seconds=59)))
                    self.start_datetime += timedelta(hours=24)
                self.date_tuple_list.append((self.start_datetime, self.end_datetime))
            for elem in self.date_tuple_list:
                print(elem)

    def __utc_to_local(self):
        tzlist = {
            'HT': 'US/Hawaii',
            'AT': 'US/Alaska',
            'PT': 'US/Pacific',
            'MT': 'US/Mountain',
            'CT': 'US/Central',
            'ET': 'US/Eastern'
        }
        print('UTC Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'Davis utilizes Unix Epoch, this routine is used for metadata purposes' + " \\ "
        self.conversion_msg += \
            'UTC start date passed as parameter: {}, local time zone: {}'.format(self.start_datetime, self.tz) + " \\ "
        self.start_datetime = None if not self.start_datetime else \
            self.start_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
        print('Local time Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'Local time start date after conversion: {}'.format(self.start_datetime) + " \\ "

        print('UTC End date: {}, local time zone: {}'.format(self.end_datetime, self.tz))
        self.conversion_msg += \
            'UTC end date passed as parameter: {}, local time zone: {}'.format(self.end_datetime, self.tz) + " \\ "
        self.end_datetime = None if not self.end_datetime else \
            self.end_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))
        print('Local time End date: {}'.format(self.end_datetime))
        self.conversion_msg += 'Local time end date after conversion: {}'.format(self.end_datetime) + " \\ "
        self.cur_datetime = self.cur_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))

    def __format_time(self):
        # this part is for the metadata purposes
        self.__utc_to_local()
        self.start_datetime = self.start_datetime.strftime('%Y-%m-%d %H:%M:%S') if self.start_datetime else None
        self.end_datetime = self.end_datetime.strftime('%Y-%m-%d %H:%M:%S') if self.end_datetime else None
        self.cur_datetime = self.cur_datetime.strftime('%Y-%m-%d %H:%M:%S')

        # this part is for actual API call
        if self.start_datetime and self.end_datetime:
            temp_list = list()
            for st, ed in self.date_tuple_list:
                st = int(st.replace(tzinfo=timezone.utc).timestamp())
                ed = int(ed.replace(tzinfo=timezone.utc).timestamp())
                temp_list.append((st, ed))
            self.date_tuple_list = temp_list

    def __compute_signature(self):

        def compute_signature_engine():  # compute_engine
            for key in params:
                print("Parameter name: \"{}\" has value \"{}\"".format(key, params[key]))

            data = ""
            for key in params:
                data = data + key + str(params[key])
            print("Data string to hash is: \"{}\"".format(data))

            sig = hmac.new(
                self.apisec.encode('utf-8'),
                data.encode('utf-8'),
                hashlib.sha256).hexdigest()
            print("API Signature is: \"{}\"".format(sig))
            return sig

        if self.start_datetime and self.end_datetime:
            temp_list = list()
            for st, ed in self.date_tuple_list:
                params = {'api-key': self.apikey,
                          'station-id': self.sn,
                          't': self.t,
                          'start-timestamp': st,
                          'end-timestamp': ed}
                params = collections.OrderedDict(sorted(params.items()))
                apisig = compute_signature_engine()
                temp_list.append((st, ed, apisig))
            self.date_tuple_list = temp_list
            for elem in self.date_tuple_list:
                print(elem)
        else:
            params = {'api-key': self.apikey, 'station-id': self.sn, 't': self.t}
            params = collections.OrderedDict(sorted(params.items()))
            self.apisig = compute_signature_engine()


class DavisReadings:
    """
    A class used to represent a device's readings
    Attributes
    ----------
    request : list
        a list of Request objects defining the request made to the Davis server
    response : list
        a (combined) raw json responses from the Davis server
    transformed_resp : list of dict
        a transformed response from raw JSON file or raw JSON response
    debug_info : dict
        a dict structure consist of parameter name and values
    """
    def __init__(self, param: DavisParam):
        """
        Gets a device readings using a GET request to the Davis API.
        Parameters
        ----------
        param : DavisParam
            DavisParam object that contains Davis API parameters
        """
        self.debug_info = {
            'sn': param.sn,
            'apikey': param.apikey,
            't': param.t,
            'apisig': param.apisig,
            'date_tuple_list': param.date_tuple_list,
            'start_datetime': param.start_datetime,
            'end_datetime': param.end_datetime,
            'cur_datetime': param.cur_datetime,
            'tz': param.tz,
            'conversion_msg': param.conversion_msg,
            'json_str': param.json_file,
            'binding_ver': param.binding_ver
        }
        self.request = list()
        self.response = list()

        if param.json_file:
            self.response = json.load(open(param.json_file))
            self.__transform()
        elif param.sn and param.apikey:
            self.__get(param.sn, param.apikey, param.apisig, param.date_tuple_list, param.t)
        elif param.sn or param.apikey:
            raise Exception('"sn" and "apikey" parameters must both be included.')
        else:
            # build an empty DavisToken
            self.request = None
            self.response = None
            self.transformed_resp = None
            # self.device_info = None
            # self.measurement_settings = None
            # self.time_settings = None
            # self.locations = None
            # self.installation_metadata = None

    def __get(self, sn, apikey, apisig, date_tuple_list, t):
        """
        Gets a device readings using a GET request to the Davis API.
        Wraps build and parse functions.
        Parameters
        ----------
        sn : str
            The serial number of the device
        apikey : str
            The user's API access key
        apisig : str
            API signature calculated when parameter object is instantiated
        date_tuple_list : list
            A list of date/time tuples with API signatures
        t : int
            Unix timestamp when the query is submitted
        """
        self.__build(sn, apikey, apisig, date_tuple_list, t)
        self.__make_request()
        self.__transform()
        return self

    def __build(self, sn, apikey, apisig, date_tuple_list: list, t):
        """
        Gets a device readings using a GET request to the Davis API.
        Parameters
        ----------
        sn : str
            The serial number of the device
        apikey : str
            The user's API access key
        apisig : str
            API signature calculated when parameter object is instantiated
        date_tuple_list : list
            A list of date/time tuples with API signatures
        t : int
            Unix timestamp when the query is submitted
        """
        if len(date_tuple_list) != 0:
            for st, ed, sig in date_tuple_list:
                tmp_request = Request('GET',
                                      url='https://api.weatherlink.com/v2/historic/' + sn,
                                      params={'api-key': apikey,
                                              't': t,
                                              'start-timestamp': st,
                                              'end-timestamp': ed,
                                              'api-signature': sig}).prepare()
                self.request.append(tmp_request)
        else:
            tmp_request = Request('GET',
                                  url='https://api.weatherlink.com/v2/current/' + sn,
                                  params={'api-key': apikey, 't': t, 'api-signature': apisig}).prepare()
            self.request.append(tmp_request)

        self.debug_info['request'] = self.request
        # self.debug_info['http_method'] = self.request.method
        # self.debug_info['url'] = self.request.url
        # self.debug_info['headers'] = self.request.headers
        return self

    def __make_request(self):
        """
        Sends a token request to the Davis API and stores the response.
        """
        # prep response list
        metadata = {
            "vendor": "davis",
            "station_id": self.debug_info['sn'],
            "timezone": self.debug_info['tz'],
            "start_datetime": self.debug_info['start_datetime'],
            "end_datetime": self.debug_info['end_datetime'],
            "request_time": self.debug_info['cur_datetime'],
            "python_binding_version": self.debug_info['binding_ver']}
        self.response.append(metadata)
        # Send the request and get the JSON response
        for req in self.request:
            resp = Session().send(req)
            if resp.status_code != 200:
                raise Exception(
                    'Request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code, resp.text))
            self.response.append(resp.json())
            # self.debug_info['response'] = self.response
        return self

    def __transform(self):
        """
        Transform the response.
        """
        def epoch_converter(epoch, time_zone):
            tzlist = {
                'HT': 'US/Hawaii',
                'AT': 'US/Alaska',
                'PT': 'US/Pacific',
                'MT': 'US/Mountain',
                'CT': 'US/Central',
                'ET': 'US/Eastern'
            }
            utc_dt = datetime.utcfromtimestamp(epoch).replace(tzinfo=pytz.utc)
            tz = pytz.timezone(tzlist[time_zone])
            dt = utc_dt.astimezone(tz)
            return dt.strftime('%Y-%m-%d %H:%M:%S')

        self.transformed_resp = list()
        station_id = self.response[0]['station_id']
        request_datetime = self.response[0]['request_time']
        for idx in range(1, len(self.response)):
            sensors = self.response[idx]['sensors']
            for jdx in range(len(sensors)):
                data = sensors[jdx]['data']
                for kdx in range(len(data)):
                    if "temp_out" not in data[kdx]:
                        break
                    else:
                        temp_dic = {
                            "station_id": station_id,
                            "request_datetime": request_datetime,
                            "data_datetime": epoch_converter(data[kdx]['ts'], self.response[0]['timezone']),
                            "atemp": round(((float(data[kdx]['temp_out']) - 32) * 5 / 9), 1),
                            "pcpn": data[kdx]['rainfall_mm'],
                            "relh": data[kdx]['hum_out']
                        }
                        self.transformed_resp.append(temp_dic)
        print(self.transformed_resp)
        return self



from datetime import datetime, timezone
import json
from time import sleep

import pytz
from requests import Session, Request


class ZentraParam:
    """
    A class used to represent Zentra API parameters
    Attributes
    ----------
    sn : str
        The serial number of the device
    token : str
        The user's access token
    start_datetime_org : datetime
        Stores datetime object passed initially
    start_datetime : datetime
        Return readings with timestamps ≥ start_time. Specify start_time in Python Datetime format
    end_datetime_org : datetime
        Stores datetime object passed initially
    end_datetime : datetime
        Return readings with timestamps ≤ end_time. Specify end_time in Python Datetime format
    conversion_msg : str
        Stores time conversion message
    tz : str
        Time zone information
    start_mrid : int, optional
        Return readings with mrid ≥ start_mrid.
    end_mrid : int, optional
        Return readings with mrid ≤ start_mrid.
    json_file : str, optional
        The path to a local json file to parse.
    binding_ver : str
        Python binding version
    """
    def __init__(self, sn=None, token=None, start_datetime=None, end_datetime=None, tz=None, start_mrid=None,
                 end_mrid=None, json_file=None, binding_ver=None):
        self.sn = sn
        self.token = token
        self.start_datetime_org = start_datetime
        self.start_datetime = start_datetime
        self.end_datetime_org = end_datetime
        self.end_datetime = end_datetime
        self.cur_datetime = datetime.now(timezone.utc)
        self.tz = tz
        self.conversion_msg = ''
        self.start_mrid = start_mrid
        self.end_mrid = end_mrid
        self.json_file = json_file
        self.binding_ver = binding_ver

        self.__check_params()
        self.__format_time()

    def __check_params(self):
        tz_option = ['HT', 'AT', 'PT', 'MT', 'CT', 'ET']
        if self.start_datetime and not isinstance(self.start_datetime, datetime):
            raise Exception('start_datetime must be datetime.datetime instance')
        if self.end_datetime and not isinstance(self.end_datetime, datetime):
            raise Exception('end_datetime must be datetime.datetime instance')
        if self.start_datetime and self.end_datetime and (self.start_datetime > self.end_datetime):
            raise Exception('start_datetime must be earlier than end_datetime')
        if not self.json_file and not (self.start_datetime and self.end_datetime):
            raise Exception('state_datetime and end_datetime must be specified')
        if self.tz and (self.tz not in tz_option):
            raise Exception('time zone options: HT, AT, PT, MT, CT, ET')
        if (self.start_datetime or self.end_datetime) and not self.tz:
            raise Exception('if start_datetime or end_datetime is specified, tz must be specified')

    def __utc_to_local(self):
        tzlist = {
            'HT': 'US/Hawaii',
            'AT': 'US/Alaska',
            'PT': 'US/Pacific',
            'MT': 'US/Mountain',
            'CT': 'US/Central',
            'ET': 'US/Eastern'
        }
        print('UTC Start date: {}, local time zone: {}'.format(self.start_datetime, self.tz))
        self.conversion_msg += \
            'UTC start date passed as parameter: {}, local time zone: {}'.format(self.start_datetime, self.tz) + " \\ "
        # self.start_datetime = self.start_datetime.replace(tzinfo=timezone.utc).astimezone(tz=None) \
        #     if self.start_datetime else None
        self.start_datetime = \
            self.start_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz])) \
            if self.start_datetime else None
        print('Local time Start date: {}'.format(self.start_datetime))
        self.conversion_msg += 'Local time start date after conversion: {}'.format(self.start_datetime) + " \\ "

        print('UTC End date: {}, local time zone: {}'.format(self.end_datetime, self.tz))
        self.conversion_msg += \
            'UTC end date passed as parameter: {}, local time zone: {}'.format(self.end_datetime, self.tz) + " \\ "
        self.end_datetime = self.end_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz])) \
            if self.end_datetime else None
        self.conversion_msg += 'Local time end date after conversion: {}'.format(self.end_datetime) + " \\ "
        print('Local time End date: {}'.format(self.end_datetime))
        self.cur_datetime = self.cur_datetime.replace(tzinfo=timezone.utc).astimezone(pytz.timezone(tzlist[self.tz]))

    def __format_time(self):
        self.__utc_to_local()
        self.start_datetime = self.start_datetime.strftime('%m-%d-%Y %H:%M') if self.start_datetime \
            else datetime.now().strftime('%m-%d-%Y %H:%M')
        self.end_datetime = self.end_datetime.strftime('%m-%d-%Y %H:%M') if self.end_datetime \
            else datetime.now().strftime('%m-%d-%Y %H:%M')
        self.cur_datetime = self.cur_datetime.strftime('%Y-%m-%d %H:%M:%S')


class ZentraReadings:
    """
    A class used to represent a device's readings
    Attributes
    ----------
    request : Request
        a Request object defining the request made to the Zentra server
    response : list
        a raw json response from the Zentra server combined with meta data
    transformed_resp : list of dict
        a transformed response from raw JSON file or raw JSON response
    debug_info : dict
        a dict structure consist of parameter name and values
    """

    def __init__(self, param: ZentraParam):
        """
        Gets a device readings using a GET request to the Zentra API.
        Parameters
        ----------
        param : ZentraParam
            ZentraParam object that contains Zentra API parameters
        """
        self.debug_info = {
            'sn': param.sn,
            'token': param.token,
            'start_datetime_org': param.start_datetime_org,
            'start_datetime': param.start_datetime,
            'end_datetime_org': param.end_datetime_org,
            'end_datetime': param.end_datetime,
            'cur_datetime': param.cur_datetime,
            'tz': param.tz,
            'conversion_msg': param.conversion_msg,
            'start_mrid': param.start_mrid,
            'end_mrid': param.end_mrid,
            'json_str': param.json_file,
            'binding_ver': param.binding_ver
        }
        if param.json_file:
            self.response = json.load(open(param.json_file))
            self.__transform()
        elif param.sn and param.token:
            self.__get(param.sn, param.token, param.start_datetime, param.end_datetime, param.start_mrid,
                       param.end_mrid)
        elif param.sn or param.token:
            raise Exception('"sn" and "token" parameters must both be included.')
        else:
            # build an empty ZentraToken
            self.request = None
            self.response = None
            self.transformed_resp = None
            # self.device_info = None
            # self.measurement_settings = None
            # self.time_settings = None
            # self.locations = None
            # self.installation_metadata = None

    def __get(self, sn, token, start_datetime=None, end_datetime=None, start_mrid=None, end_mrid=None):
        """
        Gets a device readings using a GET request to the Zentra API.
        Wraps build and parse functions.
        Parameters
        ----------
        sn : str
            The serial number of the device
        token : str
            The user's access token
        start_datetime : int, optional
            Return readings with timestamps ≥ start_datetime.
        end_datetime : int, optional
            Return readings with timestamps ≤ end_datetime.
        start_mrid : int, optional
            Return readings with mrid ≥ start_mrid.
        end_mrid : int, optional
            Return readings with mrid ≤ start_mrid.
        """
        self.__build(sn, token, start_datetime, end_datetime, start_mrid, end_mrid)
        self.__make_request()
        self.__transform()
        return self

    def __build(self, sn, token, start_datetime=None, end_datetime=None, start_mrid=None, end_mrid=None):
        """
        Gets a device readings using a GET request to the Zentra API.
        Parameters
        ----------
        sn : str
            The serial number of the device
        token : str
            The user's access token
        start_datetime : int, optional
            Return readings with timestamps ≥ start_datetime.
        end_datetime : int, optional
            Return readings with timestamps ≤ end_datetime.
        start_mrid : int, optional
            Return readings with mrid ≥ start_mrid.
        end_mrid : int, optional
            Return readings with mrid ≤ start_mrid.
        """
        self.request = Request('GET',
                               url='https://zentracloud.com/api/v3/get_readings',
                               headers={
                                   'Authorization': "Token " + token},
                               params={'device_sn': sn,
                                       'start_date': start_datetime,
                                       'end_date': end_datetime,
                                       'start_mrid': start_mrid,
                                       'end_mrid': end_mrid}).prepare()
        self.debug_info['http_method'] = self.request.method
        self.debug_info['url'] = self.request.url
        self.debug_info['headers'] = self.request.headers
        return self

    def __make_request(self):
        """
        Sends a token request to the Zentra API and stores the response.
        """
        # prep response list
        self.response = list()
        metadata = {
            "vendor": "zentra",
            "station_id": self.debug_info['sn'],
            "timezone": self.debug_info['tz'],
            "start_datetime": self.debug_info['start_datetime'],
            "end_datetime": self.debug_info['end_datetime'],
            "request_time": self.debug_info['cur_datetime'],
            "python_binding_version": self.debug_info['binding_ver']}
        self.response.append(metadata)
        # Send the request and get the JSON response
        while True:
            resp = Session().send(self.request)
            if resp.status_code != 429:
                break
            print("message: {}".format(resp.text))
            sleep(60)

        if resp.status_code != 200:
            raise Exception(
                'Request failed with \'{}\' status code and \'{}\' message.'.format(resp.status_code, resp.text))
        elif str(resp.content) == str(b'{"Error": "Device serial number entered does not exist"}'):
            raise Exception(
                'Error: Device serial number entered does not exist')
        self.response.append(resp.json())
        self.debug_info['response'] = self.response
        return self

    def __transform(self):
        """
        Parses the response.
        """
        self.transformed_resp = list()
        station_id = self.response[0]['station_id']
        request_datetime = self.response[0]['request_time']
        for idx in range(1, len(self.response)):
            try:
                data = self.response[idx]['data']
                temp_readings = data['Air Temperature'][0]["readings"]
                prec_readings = data['Precipitation'][0]["readings"]
                relh_readings = data['Relative Humidity'][0]["readings"]
            except KeyError:
                print("KeyError occurred check RAW JSON API")
                temp_dic = {
                    "station_id": station_id,
                    "request_datetime": request_datetime,
                    "data_datetime": "No Data",
                    "atemp": "No Data",
                    "pcpn": "No Data",
                    "relh": "No Data"
                }
                self.transformed_resp.append(temp_dic)
                return self
            for jdx in range(len(temp_readings)):
                temp_dic = {
                    "station_id": station_id,
                    "request_datetime": request_datetime,
                    "data_datetime": temp_readings[jdx]['datetime'][:-6],
                    "atemp": temp_readings[jdx]['value'],
                    "pcpn": prec_readings[jdx]['value'],
                    "relh": relh_readings[jdx]['value']
                }
                self.transformed_resp.append(temp_dic)
        # print(self.transformed_resp)
        return self
